import typing
import logging

logger = logging.getLogger(__name__)


def decompose_type(
    typevar: typing.Type,
    break_types: typing.List[typing.Type] = [],
    include_base_types: bool = True,
) -> typing.List[typing.Type]:
    """Decomposes a single type that takes arguments and returns a list
    containing all types referenced within, recursively.

    Arguments
    ---------
    typevar : `~typing.Type`
        type annotation to decompose.

    break_types : `~typing.List[~typing.Type]`, optional
        If a list of types is provided, instances of this particular type will
        always halt the decomposition process. So, for example, if break_types
        = [Callable], and typevar = Callable[str, float], then all_types
        (return value) = [Callable[str, float]]

    include_base_types : :py:obj:`bool`, optional
        If `False` (default `True`), base types containing other relevant types
        are ignored. These are only for *resolvable* types at return, such as
        Tuples, Lists, and Iterables. Callables are excluded by default.
    """
    # Catch non-types
    # if not isinstance(typevar, typing.Type):
    #     """Cannot decompose non-typing.Type object."""
    #     msg = f"typevar must be a type (typing.Type), not {type(typevar)}"
    #     logger.error(msg)
    #     raise TypeError(msg)

    all_types = []

    # Check that typevar is not one of interest to any of the types in
    # break_types.
    base_type = typing.get_origin(typevar)

    for bt in break_types:
        if base_type == typing.get_origin(bt):
            return [typevar]

    # If there are internal arguments (i.e., in brackets like List[str, ...])
    # then the type can be further decomposed.
    try:
        if "__args__" in typevar.__dict__:
            # Do not do this for callables, since their arguments and return
            # types are not accessible at the time they are returned.
            if (
                base_type == typing.get_origin(typing.Callable)
                or not typevar.__args__
            ):
                return [typevar]

            if include_base_types:
                all_types += [typevar]

            for arg in typevar.__args__:
                all_types += decompose_type(arg, break_types)

        else:
            all_types = [typevar]

    except AttributeError as e:
        # This should only be allowed if typevar does not have a __dict__ attr,
        # in which case we assume that this is a user-defined type.
        try:
            typevar.__dict__
            raise e

        except AttributeError:
            all_types = [typevar]

            # Log this.
            logger.debug(
                f"Return type parsed allowed without __dict__: {typevar}"
            )

    return all_types



import porchlight.param as param

from unittest import TestCase
import unittest

import itertools

import logging
import os

logging.basicConfig(filename=f"{os.getcwd()}/porchlight_unittest.log")


class ParamTest(TestCase):
    def test___init__(self):
        # Test a variety of inputs.
        names = ("bing", "bong", "bang", "sacrement", "xo xo")
        values = ("testing...", 1, 3.5, 3.0e4, tuple(), [1, 2, 3], type)
        constant = (True, False)

        for i, args in enumerate(itertools.product(names, values, constant)):
            cur_param = param.Param(*args)

            self.assertEqual(cur_param.name, args[0])
            self.assertEqual(cur_param.value, args[1])
            self.assertEqual(cur_param.constant, args[2])

    def test___eq__(self):
        param1 = param.Param("x", 1)

        param2 = param.Param("x", 1)

        self.assertEqual(param1, param2)

        param3 = param.Param("x", 2)

        self.assertNotEqual(param1, param3)

        param4 = param.Param("y", 1)

        self.assertNotEqual(param1, param4)

        param5 = param.Param("x", 0)

        self.assertNotEqual(param1, param5)

        param6 = param.Param("x", 1, True)

        self.assertNotEqual(param1, param6)

        nonparam = "bonk"

        self.assertNotEqual(param1, nonparam)

    def test___repr__(self):
        param1 = param.Param("x", 1)
        expected_string = (
            "Param(name=x, value=1, constant=False, " "type=<class 'int'>)"
        )

        self.assertEqual(repr(param1), expected_string)

    def test_properties(self):
        # Properties should not be directly mutable.
        testparam = param.Param("hi", 1)

        with self.assertRaises(AttributeError):
            testparam.type = "bing"

        with self.assertRaises(AttributeError):
            testparam.name = "bing"

        testparam.value = [1, 2, 3]

        self.assertEqual(testparam.value, [1, 2, 3])

        # This test will fail once Parameter type-checking is enabled.
        self.assertEqual(testparam.type, list)

        # Test constants
        testparam.constant = True

        with self.assertRaises(param.ParameterError):
            testparam.value = "bing"


class TestEmpty(TestCase):
    def test___eq__(self):
        empty = param.Empty()

        self.assertTrue(empty, param.Empty)
        self.assertTrue(empty, param.Empty())
        self.assertNotEqual(empty, 1)


if __name__ == "__main__":
    unittest.main()



"""Tests to verify the integrity of the porchlight.door.BaseDoor class and any
relevant helper functions or objects.
"""
from unittest import TestCase

import porchlight.door as door
from porchlight.door import BaseDoor
from porchlight.param import Empty, ParameterError, Param

import logging
import os

from typing import Callable


class TestBaseDoor(TestCase):
    def test___init__(self):
        def test_fxn(x: int) -> int:
            y = 2 * x
            return y

        door = BaseDoor(test_fxn)

        # Must contain both input and output parameter.
        arguments = ["x"]
        keyword_args = ["x"]
        return_vals = [["y"]]

        # Not comparing any values during this test.
        for arg in arguments:
            self.assertIn(arg, door.arguments)

        for kwarg in keyword_args:
            self.assertIn(kwarg, door.keyword_args)

        for retval in return_vals:
            self.assertIn(retval, door.return_vals)

        # Call the BaseDoor
        result = door(x=5)

        self.assertEqual(result, 10)

        # Decorators (with defaults) should behave as expected
        @BaseDoor
        def fxn_use_decorator(x):
            z = 2 * x
            return z

        self.assertEqual(fxn_use_decorator(10), 20)

        self.assertEqual(fxn_use_decorator.name, "fxn_use_decorator")

        self.assertEqual(fxn_use_decorator.arguments, {"x": Empty})

        # Test on a decorated function.
        def test_decorator(fxn):
            def wrapper(*args, **kwargs):
                return fxn(*args, **kwargs)

            return wrapper

        @test_decorator
        def test_fxn(x: int) -> int:
            y = 2 * x
            return y

        door = BaseDoor(test_fxn)

        # Must contain both input and output parameter.
        arguments = ["x"]
        keyword_args = ["x"]
        return_vals = [["y"]]

        # Not comparing any values during this test.
        for arg in arguments:
            self.assertIn(arg, door.arguments)

        for kwarg in keyword_args:
            self.assertIn(kwarg, door.keyword_args)

        for retval in return_vals:
            self.assertIn(retval, door.return_vals)

        # Call the BaseDoor
        result = door(x=5)

        self.assertEqual(result, 10)

    def test___call__(self):
        def test_fxn(x: int) -> int:
            y = 2 * x
            return y

        door = BaseDoor(test_fxn)

        # Call the BaseDoor with erroneous types based on annotations.
        with self.assertRaises(ParameterError):
            door(x="6")

    def test__get_return_vals(self):
        # Should be able to handle one return val.
        def fxn_one_return(x):
            y = 1
            return y

        result = BaseDoor._get_return_vals(fxn_one_return)

        self.assertEqual(result, [["y"]])

        # Multiple return values, no inputs
        def fxn_no_inputs():
            x, y, z, = (
                0,
                0,
                0,
            )
            return x, y, z

        result = BaseDoor._get_return_vals(fxn_no_inputs)

        self.assertEqual([["x", "y", "z"]], result)

        # No-return function.
        def fxn_no_return(x, y):
            pass

        result = BaseDoor._get_return_vals(fxn_no_return)

        self.assertEqual([], result)

        # Specified return types.
        def fxn_return_types(x: int, y) -> str:
            outstr = f"{x} || {y}"
            return outstr

        result = BaseDoor._get_return_vals(fxn_return_types)

        self.assertEqual([["outstr"]], result)

        # Test retrieving a return value for a function that wraps another
        # function to return.
        def test_wrap():
            def int_func():
                output = "Hello world!"
                return output

            return int_func

        result = BaseDoor._get_return_vals(test_wrap)

        self.assertEqual(result, [["int_func"]])

        # Test single-line definitions.
        def test_oneline(x):
            return x

        result = BaseDoor._get_return_vals(test_oneline)

        self.assertEqual(result, [["x"]])

        def test_oneline_bad(x):
            return x * 2

        result = BaseDoor._get_return_vals(test_oneline_bad)

        self.assertEqual(result, [])

        # Tests with non-return local variables.
        class Return_Type:
            pass

        def bad_ret_vals() -> Return_Type:
            return_at_beginning = None
            mid_return_loc = None
            end_loc_return = None

            # To appease flake8, but also a test itself.
            other_ret = (return_at_beginning, mid_return_loc, end_loc_return)
            other_ret = "Putting return in a new place: return"

            other_ret = Return_Type()
            return other_ret

        result = BaseDoor._get_return_vals(bad_ret_vals)

        self.assertEqual(result, [["other_ret"]])

        # Test internal decorators.
        def int_decorator(x: int) -> int:
            @BaseDoor
            def int_func():
                return 10

            y = x + int_func()
            return y

        result = BaseDoor._get_return_vals(int_decorator)

        self.assertEqual(result, [["y"]])

        # Test decorators.
        def dummy_decorator(fun) -> Callable:
            def wrapper(*args, **kwargs):
                return fun(*args, **kwargs)

            return wrapper

        @dummy_decorator
        def test_decorator() -> int:
            x = 1
            return x

        result = BaseDoor._get_return_vals(test_decorator)

        self.assertEqual(result, [["x"]])

    def test___eq__(self):
        @BaseDoor
        def fxn1(x, y):
            x += y

            return x

        @BaseDoor
        def fxn2(x, y):
            return y

        other_fxn1 = door.BaseDoor(fxn1._base_function)

        self.assertTrue(fxn1 == fxn1)
        self.assertFalse(fxn1 == fxn2)
        self.assertTrue(fxn1 == other_fxn1)

    def test___repr__(self):
        """Test the BaseDoor"""

        def test(x: int) -> int:
            y = x ** 2
            return y

        expected_repr = (
            f"BaseDoor(name=test, base_function={str(test)}, "
            f"arguments={{'x': <class 'int'>}}, "
            f"return_vals=[['y']])"
        )

        test_door = BaseDoor(test)
        self.assertEqual(repr(test_door), expected_repr)

    def test__inspect_base_callable(self):
        def bigfxn(x, y, *, z=5):
            output = sum((x, y, z))
            return output

        new_door = BaseDoor(bigfxn)

        self.assertEqual(new_door.keyword_only_args, {"z": Param("z", 5)})

        # For now, there should *not* be any functionality for required
        # positional arguments. That will come later, and this test will need
        # to be updated.
        def bad_fxn(x1, y1, /, x3, *, z7):
            c = "blah"
            return c

        with self.assertRaises(NotImplementedError):
            BaseDoor(bad_fxn)

    def test__returned_def_to_door(self):
        def my_func_gen():
            def int_func():
                output = "bingbong"
                return output

            return int_func

        no_int_ret_door = BaseDoor(my_func_gen)
        int_ret_door = BaseDoor(my_func_gen, returned_def_to_door=True)

        self.assertEqual(no_int_ret_door.name, int_ret_door.name)
        self.assertEqual(no_int_ret_door.arguments, int_ret_door.arguments)

        self.assertTrue(isinstance(no_int_ret_door(), Callable))
        self.assertTrue(isinstance(int_ret_door(), BaseDoor))

    def test_keyword_arguments(self):
        @BaseDoor
        def test_prop(arg1, arg2, kwarg1=None, kwarg2=["hi"]) -> int:
            kwarg1 = 1 if kwarg1 is None else kwarg1 + 1
            return kwarg1

        expected_val = {
            "arg1": Param("arg1", Empty()),
            "arg2": Param("arg2", Empty()),
            "kwarg1": Param("kwarg1", None),
            "kwarg2": Param("kwarg2", ["hi"]),
        }

        self.assertEqual(test_prop.keyword_arguments, expected_val)

    def test_kwargs(self):
        @BaseDoor
        def test_prop(arg1, arg2, kwarg1=None, kwarg2=["hi"]) -> int:
            kwarg1 = 1 if kwarg1 is None else kwarg1 + 1
            return kwarg1

        expected_val = {
            "arg1": Param("arg1", Empty()),
            "arg2": Param("arg2", Empty()),
            "kwarg1": Param("kwarg1", None),
            "kwarg2": Param("kwarg2", ["hi"]),
        }

        self.assertEqual(test_prop.kwargs, expected_val)


if __name__ == "__main__":
    import unittest

    unittest.main()



import inspect
import re

from .param import Empty, ParameterError, Param
from .utils.typing_functions import decompose_type
from .utils.inspect_functions import get_all_source, get_wrapped_function

import typing
from typing import Any, Callable, Dict, List, Type

import logging

logger = logging.getLogger(__name__)


class BaseDoor:
    """Contains the basic information of a Door object, as well as base
    functions.

    Attributes
    ----------
    arguments : :py:obj:`dict`, :py:obj:`str`: :class:`~typing.Type`
        Dictionary of all arguments taken as input when the `BaseDoor` object
        is called.

    keyword_args : :py:obj:`dict`, :py:obj:`str`: :class:`~typing.Any`
        Keyword arguments accepted by the `BaseDoor` as input when called. This
        includes all arguments that are not positional-only. Positional
        arguments without a default value are assigned a
        :class:~porchlight.param.Empty` value instead of their default value.

    max_n_return : :py:obj:`int`
        Maximum number of returned values.

    min_n_return : :py:obj:`int`
        Minimum number of returned values.

    n_args : :py:obj:`int`
        Number of arguments accepted by this `BaseDoor`

    name : :py:obj:`str`
        The name of the function as visible from the base function's __name__.

    return_types : :py:obj:`list` of :py:obj:`list` of `~typing.Type`
        Values returned by any return statements in the base function.

    return_vals : :py:obj:`list` of :py:obj:`list` of :py:obj:`str`
        Values returned by any return statements in the base function.

    typecheck : :py:obj:`bool`
        If True, when arguments are passed to the `BaseDoor`'s base function
        the input types are checked against the types in `BaseDoor.arguments`.
        If there is a mismatch, a `TypeError` will be thrown.

    _base_function : :py:obj:`~typing.Callable`
        This holds a reference to the function being managed by the `BaseDoor`
        instance.
    """

    _base_function: Callable
    arguments: Dict[str, Type]
    keyword_args: Dict[str, Any]
    max_n_return: int
    min_n_return: int
    n_args: int
    name: str
    return_types: List[List[Type]]
    return_vals: List[List[str]]
    typecheck: bool

    def __init__(
        self,
        function: Callable,
        typecheck: bool = True,
        returned_def_to_door: bool = False,
    ):
        """Initializes the BaseDoor class. It takes any callable (function,
        lambda, method...) and inspects it to get at its arguments and
        structure.

        if typecheck is True (default True), the type of inputs passed to
        BaseDoor.__call__ will be checked for matches to known input Types.

        Parameters
        ----------
        function : :py:class:`typing.Callable`
            The callable/function to be managed by the BaseDoor class.

        typecheck : :py:obj:`bool`, optional
            If `True`, the `BaseDoor` object will assert that arguments passed
            to `__call__` (when the `BaseDoor` itself is called like a
            function) have the type expected by type annotations and any user
            specifications. By default, this is `True`.

        returned_def_to_door : :py:obj:`bool`, optional
            Returns a Door generated from the output of the base function.
            Note, this is not the same as a DynamicDoor, and internal
            variables/updating is not handled as with a DynamicDoor. This just
            calls Door's initializer on the output of the base function.
        """
        self._returned_def_to_door = returned_def_to_door
        self._base_function = function
        self.typecheck = typecheck
        self._inspect_base_callable()

    def __eq__(self, other) -> bool:
        """Equality is defined as referencing the same base function."""
        if self.name is other.name:
            return True

        return False

    def __repr__(self):
        info = {
            "name": self.name,
            "base_function": self._base_function,
            "arguments": self.arguments,
            "return_vals": self.return_vals,
        }

        substrs = [f"{key}={value}" for key, value in info.items()]

        outstr = f"BaseDoor({', '.join(substrs)})"

        return outstr

    def _inspect_base_callable(self):
        """Inspect the BaseDoor's baseline callable for primary attributes.

        This checks for type annotations, return statements, and all
        information accessible to :py:obj:`inspect.Signature` relevant to
        `BaseDoor`.
        """
        # Need to find the un-wrapped function that actually takes the
        # arguments in the end.
        function = get_wrapped_function(self._base_function)

        self.name = function.__name__
        self.__name__ = function.__name__
        self.arguments = {}
        self.keyword_args = {}
        self.keyword_only_args = {}

        try:
            ret_type_annotation = typing.get_type_hints(function)["return"]
            self.return_types = decompose_type(
                ret_type_annotation, include_base_types=False
            )

        except KeyError:
            # No return types there.
            self.return_types = None

        for name, param in inspect.signature(function).parameters.items():
            self.arguments[name] = param.annotation

            # Check for positional-only arguments first, then proceed to other
            # argument types.
            if param.kind == inspect.Parameter.POSITIONAL_ONLY:
                msg = (
                    f"porchlight does not support positional-only "
                    f"arguments, which were found in {self.name}."
                )

                logger.error(msg)
                raise NotImplementedError(msg)

            elif param.default != inspect._empty:
                self.keyword_args[name] = Param(name, param.default)

                if param.kind == inspect.Parameter.KEYWORD_ONLY:
                    self.keyword_only_args[name] = Param(name, param.default)

            elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                # This catches keyword-only arguments with no default value.
                # ^ TKREFACTORING is that a thing we need to worry about?
                self.keyword_args[name] = Param(name, Empty())
                self.keyword_only_args[name] = Param(name, Empty())

            else:
                self.keyword_args[name] = Param(name, Empty())

        for name, _type in self.arguments.items():
            if _type == inspect._empty:
                self.arguments[name] = Empty

        self.n_args = len(self.arguments)

        # The return values require some more effort.
        self.return_vals = self._get_return_vals(function)

        logger.debug(f"Found {self.n_args} arguments in {self.name}.")

    def __call__(self, *args, **kwargs):
        """Calls the BaseDoor's function as normal.

        The keys of kwargs must be passed with the same name as variables
        within the BaseDoor. As of right now, if extra values are included,
        they are ignored.
        """
        # This is currently filtering out all the arguments the function
        # actually needs. But this way of doing it should probably be handled
        # by the Neighborhood object in the future.
        # ^ TK REFACTORING
        input_kwargs = {k: v for k, v in kwargs.items() if k in self.arguments}

        # Type checking.
        if self.typecheck:
            for k, v in input_kwargs.items():
                if self.arguments[k] == Empty:
                    continue

                if not isinstance(v, self.arguments[k]):
                    msg = (
                        f"Type checking is on, and the type for input "
                        f"parameter {k}, {type(v)} to {self.name} does not "
                        f"match the detected type(s), {self.arguments}"
                    )

                    logger.error(msg)
                    raise ParameterError(msg)

        return_val = self._base_function(*args, **input_kwargs)

        if self._returned_def_to_door:
            if not isinstance(return_val, BaseDoor):
                return_val = BaseDoor(return_val)

        return return_val

    @staticmethod
    def _get_return_vals(function: Callable) -> List[str]:
        """Gets the names of the return value variables for a given
        function.

        For functions that wrap other functions (i.e., the return value is a
        function def'd in the function body), the return statements in the body
        are not parsed.

        Arguments
        ---------
        function : :py:class:`typing.Callable`
            The function to retrieve the return values for.
        """
        return_vals = []

        lines, start_line = get_all_source(function)

        # Tracking indentation for python-like parsing.
        cur_indent = 0
        last_check_indent = 0
        checking_for_returns = True
        main_def_found = False

        # These include mandatory space at the beginning since syntactically
        # there must exist non-\n whitespace for all lines after a funciton
        # definition.
        defmatch_str = r"^(\ )+def\s+"
        retmatch_str = r".*\s+return\s(.*)"
        indentmatch_str = r"^(\s)*"

        for i, line in enumerate(lines):
            orig_line = line.strip()

            # Remove comments
            if "#" in line:
                line = line[: line.index("#")]

            # Ignore empty lines
            if not line.strip():
                continue

            # Get the current indent level.
            indentmatch = re.match(indentmatch_str, line)
            cur_indent = len(indentmatch.group())

            # Ignore empty lines
            # Check for matches for both, in case there's something like
            #   def wtf(): return 5
            # Which is atrocious but possible.
            defmatch = re.match(defmatch_str, line)
            retmatch = re.match(retmatch_str, line)

            # Ignore decorators
            if re.match(r"\s*@\w+.*", line):
                continue

            # Catch in-function definitions and ignore them.
            if defmatch and i > 0 and main_def_found:
                checking_for_returns = False
                last_check_indent = cur_indent
                continue

            if not main_def_found and defmatch:
                main_def_found = True

            if not checking_for_returns and cur_indent <= last_check_indent:
                # This is outside the def scope
                checking_for_returns = True
                last_check_indent = 0

            if not checking_for_returns:
                continue

            if retmatch:
                # This is a return value that belongs to the object.
                ret_string = retmatch.group(1)
                if "," in ret_string:
                    vals = ret_string.split(",")

                else:
                    vals = [ret_string]

                vals = [v.strip() for v in vals]

                for val in vals:
                    if not re.match(r"\w+$", val):
                        # This is undefined, not an error. So assign return
                        # value 'undefined' for this return statement and issue
                        # a warning.
                        source_file = inspect.getfile(function)
                        logger.warning(
                            f"Could not define any set of return variable "
                            f"names for the following return line: \n"
                            f"{source_file}: {start_line+i}) "
                            f"{orig_line.strip()}\n While not crucial to "
                            f"this function, be aware that this means no "
                            f"return value will be modified by this "
                            f"callable."
                        )

                        vals = []
                        break

                if vals:
                    return_vals.append(vals)

        return return_vals

    @property
    def keyword_arguments(self):
        return self.keyword_args

    @property
    def kwargs(self):
        return self.keyword_args


class Door(BaseDoor):
    """Inherits from and extends :class:`~porchlight.door.BaseDoor`"""

    def __init__(self, function: Callable):
        super().__init__(function)

    def __repr__(self):
        return super().__repr__().replace("BaseDoor", "Door")

    @property
    def variables(self) -> List[str]:
        """Returns a list of all known return values and input arguments."""
        all_vars = []

        for arg in self.arguments:
            if arg not in all_vars:
                all_vars.append(arg)

        for ret in self.return_vals:
            for r in ret:
                if r not in all_vars:
                    all_vars.append(r)

        return all_vars

    @property
    def required_arguments(self) -> List[str]:
        """Returns a list of arguments with no default values."""
        required = []

        for x in self.arguments:
            if isinstance(self.keyword_args[x].value, Empty):
                required.append(x)

        return required


class DynamicDoor(Door):
    """A dynamic door takes a door-generating function as its initializer.

    Unlike `~porchlight.door.BaseDoor`s and `~porchlight.door.Door`s, dynamic
    doors will only parse the definition's source once it is generated.

    These objects a function that returns a Door. The `DynamicDoor` then
    contains identical attributes to the generated door. Once called again, all
    attributes update to match the most recent call.

    Attributes
    ----------
    _door_generator : `Callable`
        A function returning a `~porchlight.door.Door` as its output.

    generator_args : `List`
        List of arguments to be passed as positional arguments to the
        `_door_generator` function.

    generator_kwards : `Dict`
        Dict of key: value pairs representing keyword arguments to be passed as
        positional arguments to the `_door_generator` function.
    """

    def __init__(
        self,
        door_generator: Callable[Any, Door],
        generator_args: List = [],
        generator_kwargs: Dict = {},
    ):
        """Initializes the Dynamic Door. When __call__ is invoked, the door
        generator is called.

        Arguments
        ---------
        _door_generator : `~typing.Callable[Any, Door]`
            A callable function that returns an initialized Door object.

        generator_args : `~typing.List`
            List of positional arguments for the generator function.

        generator_kwargs : `~typing.Dict[str, Any]`
            Dictionary of `str`/`Any` key/value pairs for keyword arguments
            passed to the generator function.
        """
        self._door_generator = door_generator
        self.generator_args = generator_args
        self.generator_kwargs = generator_kwargs

        # In order to pass some checks, a few key attributes need to be
        # initialized. These will be initialized to whatever value evaluates to
        # False for the relevant type.
        self.name = ""
        self.__name__ = ""
        self.arguments = {}
        self.keyword_args = {}
        self.keyword_only_args = {}
        self.return_vals = []
        self._cur_door = None
        self._last_door = None

    def __call__(self, *args, **kwargs) -> Any:
        """Executes the function stored in
        `~porchlight.door.DynamicDoorc._base_function` once
        `~porchlight.door.DynamicDoor.update` has executed.

        Arguments
        ---------
        *args : positional arguments
            Arguments directly passed to
            `~porchlight.door.DynamicDoor._base_function`.

        **kwargs : keyword arguments
            Keyword arguments directly passed to
            `~porchlight.door.DynamicDoor._base_function`.
        """
        self.update()
        result = self._base_function(*args, **kwargs)

        return result

    def call_without_update(self, *args, **kwargs) -> Any:
        """Executes the function stored in
        `~porchlight.door.DynamicDoor._base_function` *WITHOUT* executing
        `~porchlight.door.DynamicDoor.update()`

        Arguments
        ---------
        *args : positional arguments
            Arguments directly passed to
            `~porchlight.door.DynamicDoor._base_function`.

        **kwargs : keyword arguments
            Keyword arguments directly passed to
            `~porchlight.door.DynamicDoor._base_function`.
        """
        # Give a specific, useful message if self._base_function is not
        # initialized.
        if "_base_function" not in self.__dict__:
            msg = (
                "DynamicDoor does not contain _base_function attribute. "
                "Did you means to call DynamicDoor directly or with "
                "DirectDoor()?"
            )

            logger.error(msg)
            raise AttributeError(msg)

        result = self._base_function(*args, **kwargs)

        return result

    def __repr__(self):
        # Try generating a __str__ using the BaseDoor.__str__ method. If it
        # fails, then revert to an uninitialized DynamicDoor string as
        # appropriate.
        try:
            outstr = super().__repr__().replace("Door", "DynamicDoor")

        except AttributeError:
            outstr = "DynamicDoor(uninitialized)"

        return outstr

    def update(self):
        """Updates the DynamicDoor using `DynamicDoor._door_generator`

        This method is called when DynamicDoor.__call__ is invoked.
        """
        self._last_door = self._cur_door

        updated_door = self._door_generator(
            *self.generator_args, **self.generator_kwargs
        )

        if not isinstance(updated_door, BaseDoor):
            msg = f"Expected a BaseDoor object, but {updated_door} returned."
            logger.error(msg)
            raise TypeError(msg)

        # This is so incredibly sus... but it works for now. There is
        # definitely a more elegant way to do this.
        for attr, value in updated_door.__dict__.items():
            # Ignore all dunder attrs.
            if attr[:2] == "__":
                continue

            setattr(self, attr, value)

        self._cur_door = updated_door



"""Defines the `Neighborhood` class."""
from . import door
from . import param
from .utils.typing_functions import decompose_type
from typing import Any, Callable, Dict, List, Union

import logging

logger = logging.getLogger(__name__)


class NeighborhoodError(Exception):
    pass


class Neighborhood:
    """A neighborhood manages the interactions between Doors. It consists of a
    modifiable collection of :class:`~porchlight.door.Door` (or
    :class:`~porchlight.door.BaseDoor`) objects.

    Attributes
    ----------
    _doors : :py:obj:`dict`, :py:obj:`str`: :class:`~porchlight.door.Door`
        Contains all data for :class:`~porchlight.door.Door` objects. The keys
        are, by default, the :meth:`~porchlight.door.Door.name`
        property for the corresponding :class:`~porchlight.door.Door` values.

    _params : :py:obj:`dict`, :py:obj:`str`: :class:`~porchlight.param.Param`
        Contains all the parameters currently known to and managed by the
        :class:`~porchlight.neighborhood.Neighborhood` object.

    _call_order : :py:obj:`list`, :py:obj:`str`
        The order in which the :class:`~porchlight.door.Door` objects in
        `_doors` are called. By default, this is the order in which
        :class:`~porchlight.door.Door`s are added to the `Neighborhood`.
    """

    _doors: Dict[str, door.Door]
    _dynamic_doors: set[str]
    _params: Dict[str, param.Param]
    _call_order: List[str]

    def __init__(self):
        """Initializes the Neighborhood object."""
        self._doors = {}
        self._params = {}
        self._call_order = []

    def __repr__(self):
        """Must communicate teh following:
        + A unique identifier.
        + List of doors
        + list of tracked parameters and their values.
        """
        info = {
            "doors": self.doors,
            "params": self.params,
            "call_order": self._call_order,
        }

        infostrs = [f"{key}={value}" for key, value in info.items()]

        outstring = f"Neighborhood({', '.join(infostrs)})"

        return outstring

    def add_function(
        self,
        function: Callable,
        overwrite_defaults: bool = False,
        dynamic_door: bool = False,
    ):
        """Adds a new function to the Neighborhood object.

        Parameters
        ----------
        function : Callable
            The function to be added. This is converted to a
            :class:`~porchlight.door.Door` object by this method.

        overwrite_defaults : bool, optional
            If `True`, will overwrite any parameters shared between the
            `function` and `Neighborhood._params` to be equal to the defaults
            set by `function`. If `False` (default), no parameters that exist
            in the `Neighborhood` object already will be changed.

        dynamic_door : bool, optional
            If `True` (default `False`), then the output(s) of this `Door` will
            be converted into a `Door` or set of `Door`s in the `Neighborhood`
            object.
        """
        new_door = door.Door(function)

        self.add_door(new_door, overwrite_defaults, dynamic_door)

    def add_door(
        self,
        new_door: Union[door.Door, List[door.Door]],
        overwrite_defaults: bool = False,
        dynamic_door: bool = False,
    ):
        """Adds an already-initialized Door to the neighborhood.

        Parameters
        ----------
        new_door : :class:`~porchlight.door.Door`,
            :class:`~porchlight.door.DynamicDoor`, or :py:obj:`list` of
            :class:`~porchlight.door.Door` objects.

            Either a single initialized
            `door.Door` object or a list of them.  If a list is provided, this
            function is called for each item in the list.

        overwrite_defaults : bool, optional
            If `True`, will overwrite any parameters shared between the
            `new_door` and `Neighborhood._params` to be equal to the defaults
            set by `new_door`. If `False` (default), no parameters that exist
            in the `Neighborhood` object already will be changed.

        dynamic_door : bool, optional
            If `True` (default `False`), then the output(s) of this `Door` will
            be converted into a `Door` or set of `Door`s in the `Neighborhood`
            object.
        """
        if isinstance(new_door, List):
            for nd in new_door:
                self.add_door(
                    nd,
                    overwrite_defaults=overwrite_defaults,
                    dynamic_door=dynamic_door,
                )

            return

        self._doors[new_door.name] = new_door

        # Add the new door to the call order.
        self._call_order.append(new_door.name)

        # Update the parameters as necesssary.
        # This will prefer default values passed by earlier doors.
        for pname, ptype in new_door.arguments.items():
            if pname not in self._params or overwrite_defaults:
                # This is a new parameter.
                parameter_name = pname
                value = param.Empty()

                if pname in new_door.keyword_args:
                    value = new_door.keyword_args[pname]

                self.add_param(parameter_name, value)

        # If not well-defined, we cannot modify outputs to this function.
        if not new_door.return_vals:
            return

        # Add all return values as parameters.
        if not dynamic_door:
            for pname in [p for rvs in new_door.return_vals for p in rvs]:
                if pname not in self._params:
                    self._params[pname] = param.Param(pname, param.Empty())

            return

        # Dynamic doors must be specified separately. They get initialized when
        # first modified.
        #
        # Dynamic doors must also be type-annotated. If they are not raise an
        # error.
        if not new_door.return_types:
            msg = (
                "DynamicDoor requires a type annotation for the "
                "function generator."
            )

            logger.error(msg)
            raise NeighborhoodError(msg)

        return_types = new_door.return_types

        if not return_types or len(return_types) != len(
            new_door.return_vals[0]
        ):
            raise NeighborhoodError(
                "Dynamic doors with return values must be type annotated."
            )

        for i, rt in enumerate(return_types):
            if isinstance(rt, door.Door) or rt is door.Door:
                ret_val = new_door.return_vals[0][i]
                if ret_val not in self.doors:
                    # Can define a function that just pulls out the attr
                    # independent of its current reference.
                    @door.DynamicDoor
                    def template_ddoor(param_name):
                        return getattr(self, "params")[param_name].value

                    # Need to rename this appropriately for the neighborhood.
                    # And provide a static copy of the current return value's
                    # name.
                    template_ddoor.name = ret_val
                    template_ddoor.generator_kwargs = {"param_name": ret_val}

                    self.add_door(template_ddoor)
                    self.add_param(ret_val, param.Empty)

    def remove_door(self, name: str):
        """Removes a :class:`~porchlight.door.Door` from :attr:`_doors`.

        Parameters
        ----------
        name : :py:obj:`str`
            The name of the :class:`~porchlight.door.Door` to be removed. It
            must correspond to a key in `Neighborhood._doors` attribute.

        Raises
        ------
        KeyError
            If `name` is not present in `_doors` attribute.
        """
        if name not in self._doors:
            raise KeyError(f"Could not find a door named {name}.")

        del self._doors[name]

    def set_param(
        self,
        parameter_name: str,
        new_value: Any,
        constant: bool = False,
        *,
        ignore_constant: bool = False,
    ) -> param.Param:
        """Set the value of a parameter to a new value. Since parameters are
        not meant to be modified like this right now, generate a new parameter.

        Parameters
        ----------
        parameter_name : :py:obj:`str`
            The name of the parameter to modify.

        new_value : `Any`
            The value to be assigned to this parameter.

        constant : :py:obj:`bool`
            Will be passed to :class:`~porchlight.param.Param` as the
            `constant` keyword argument.

        ignore_constant : :py:obj:`bool`, optional, keyword-only
            When assigning this parameter, it will ignore the `constant`
            attribute of the current parameter.

        Raises
        ------
        :class:`~porchlight.param.ParameterError`
            Is raised if the parameter attempting to be changed has `True` for
            its `constant` attribute. Will not be raised by this method if
            `ignore_constant` is `True`.
        """
        _old_param = self._params[parameter_name]

        if not ignore_constant and _old_param.constant:
            msg = f"Parameter {_old_param.name} is set to constant."

            logger.error(msg)

            raise param.ParameterError(msg)

        self._params[parameter_name] = param.Param(
            parameter_name, new_value, constant
        )

    def add_param(
        self, parameter_name: str, value: Any, constant: bool = False
    ):
        """Adds a new parameter to the `Neighborhood` object.

        The parameters all correspond to arguments passed directly to the
        :class:`~porchlight.param.Param` initializer.

        Parameters
        ----------
        parameter_name : :py:obj:`str`
            Name of the parameter being created.

        value : `Any`
            Parameter value

        constant : :py:obj:`bool`, optional
            If `True`, the parameter is set to constant.
        """
        if not isinstance(value, param.Param):
            new_param = param.Param(parameter_name, value, constant)

        else:
            new_param = value

        self._params[parameter_name] = new_param

    def required_args_present(self) -> bool:
        """Returns True if all the necessary arguments to run all Doors in the
        Neighborhood object are included in the Neighborhood object.
        """
        done = []
        for d_params in (d.arguments for d in self.doors.values()):
            for pname in d_params:
                if pname in done:
                    continue

                done.append(pname)

                pname_missing = pname not in self._params
                is_empty_param = False

                if not pname_missing:
                    is_empty_param = isinstance(
                        self._params[pname].value, param.Empty
                    )

                if pname_missing or is_empty_param:
                    return False

        return True

    def call_all_doors(self):
        """Calls every door currently present in the neighborhood object.

        This order is currently dictated by the order in which
        :class:`~porchlight.door.Door`s are added to the `Neighborhood`.

        The way this is currently set up, it will not handle positional
        arguments. That is, if an input cannot be passed using its variable
        name, this will break.
        """
        # Need to call the doors directly.
        for doorname in self._call_order:
            # Dynamic doors must first be updated and parsed.
            if isinstance(self._doors[doorname], door.DynamicDoor):
                self._doors[doorname].update()

                # Need to ensure all params are present in the neighborhood.
                dynamic_params = self._doors[doorname].variables

                for p in (p for p in dynamic_params if p not in self._params):
                    # Check for a default value here.
                    if p in self._doors[doorname].keyword_args:
                        value = self._doors[doorname].keyword_args[p]

                    else:
                        value = param.Empty

                    self.add_param(p, value)

            cur_door = self._doors[doorname]
            req_params = cur_door.arguments
            input_params = {}

            for pname in req_params:
                input_params[pname] = self._params[pname].value

            # Run the cur_door object and catch its output.
            output = cur_door(**input_params)

            # Check if the cur_door has a known return value.
            if not cur_door.return_vals:
                continue

            elif len(cur_door.return_vals[0]) > 1:
                # This only works for functions with one possible output. This,
                # frankly, should probably be the case nearly all of the time.
                # Still need to make a call on if there's support in a subset
                # of cases. See issue #19 for updates.
                update_params = {
                    v: x for v, x in zip(cur_door.return_vals[0], output)
                }

            else:
                assertmsg = "Mismatched output/return."
                assert len(cur_door.return_vals) == 1, assertmsg
                update_params = {cur_door.return_vals[0][0]: output}

            for pname, new_value in update_params.items():
                # If the parameter is currently empty, just reassign and
                # continue. This refreshes the type value of the parameter
                if isinstance(self._params[pname].value, param.Empty):
                    self._params[pname] = param.Param(pname, new_value)

                # If the parameter is supposed to be constant, raise an error.
                if pname in self._params and self._params[pname].constant:
                    msg = (
                        f"Door {cur_door.name} is attempting to change a "
                        f"constant parameter: {pname}."
                    )

                    logger.error(msg)
                    raise param.ParameterError(msg)

                # Editing the _value directly here... but I'm not sure if
                # that's the best idea.
                self._params[pname]._value = new_value

    def run_step(self):
        """Runs a single step forward for all functions, in specified order,
        based on the current parameter state of the Neighborhood object.

        The way this is currently set up, it will not handle positional
        arguments. That is, if an input cannot be passed using its variable
        name, this will break.
        """
        # Ensure that all required parameters are defined.
        empty_params = self.empty_parameters
        req_args = self.required_parameters

        if any(e in req_args for e in empty_params):
            missing = [e for e in empty_params if e in req_args]
            msg = f"Missing parameters required for input: {missing}."
            logger.error(msg)
            raise param.ParameterError(msg)

        self.call_all_doors()

    def order_doors(self, order: List[str]):
        """Allows the doors to be ordered when called.

        If this is never called, the call order will be equivalent to the order
        in which the doors are added to the `Neighborhood`. As of right now,
        all doors present must be included in the `order` argument or a
        `KeyError` will be thrown.

        Arguments
        ---------
        order : :py:obj:`list`, str
            The order for doors to be called in. Each `str` must correspond to
            a key in `Neighborhood._doors`.
        """
        if not order:
            msg = f"Empty or invalid input: {order}."
            logger.error(msg)
            raise ValueError(msg)

        elif len(order) > len(self._doors):
            msg = (
                f"More labels provided than doors "
                f"({len(order)} labels vs {len(self._doors)} doors)."
            )
            logger.error(msg)
            raise ValueError(msg)

        elif any(n not in self._doors for n in order):
            i = [n not in self._doors for n in order].index(True)

            msg = f"Could not find door with label: {order[i]}"
            logger.error(msg)
            raise KeyError(msg)

        self._call_order = order

    @property
    def doors(self):
        return self._doors

    @property
    def empty_parameters(self):
        empty_params = {}

        for pname, p in self._params.items():
            if isinstance(p.value, param.Empty):
                empty_params[pname] = p

        return empty_params

    @property
    def parameters(self):
        return self._params

    @property
    def params(self):
        return self._params

    @property
    def required_parameters(self):
        """Defines parameters that must not be empty at the start of a run
        for the run to successfully execute.
        """
        req_args = set()
        returned_args = set()

        for d in self._doors.values():
            for pname in d.required_arguments:
                if pname not in returned_args:
                    req_args.add(pname)

                # Add returned values that would now be initialized. This is
                # assuming that all returns will successfully.
                for ra in d.return_vals:
                    for p in ra:
                        returned_args.add(p)

        return req_args

    @property
    def uninitialized_inputs(self):
        """Parameters that are inputs to Doors within this Neighborhood with
        Empty values.
        """
        input_args = []

        for d in self._doors.values():
            for pname in d.required_arguments:
                if (
                    pname not in input_args
                    and self._params[pname].value == param.Empty()
                ):
                    input_args.append(pname)

        return input_args



"""Unit testing for porchlight/utils/inspect_functions."""
import porchlight.utils.inspect_functions as inspect_functions
import inspect
import unittest


def _helper_indent_and_newline(
    lines: list[str], indent: int, indent_blank: bool = True
):
    """Takes a list of strings and returns those strings."""
    for i, line in enumerate(lines):
        if line:
            line = " " * indent + line

        if i < len(lines):
            line += "\n"

        lines[i] = line

    return lines


class TestInspectFunctions(unittest.TestCase):
    def test_get_all_source(self):
        # The baseline case should reduce directly to inspec.getsourcelines
        def test1(bing_bong: bool = True) -> str:
            """This is my internal docstring."""
            switch = bing_bong

            # This is a comment line.
            result = "The switch is "
            result += "on" if switch else "off"

            return result

        expected_result = inspect.getsourcelines(test1)

        result = inspect_functions.get_all_source(test1)

        self.assertEqual(result[0], expected_result[0])
        self.assertEqual(result[1], expected_result[1])

        # Decorator handling.
        def test2_decorator(fxn):
            """This is a dummy wrapper."""

            def wrapper(*args, **kwargs):
                result = fxn(*args, **kwargs)
                return result

            return wrapper

        @test2_decorator
        def test2(a, b, c):
            """A docstring for this test function."""
            # A comment in this test function.
            total = sum(a, b, c)
            outstr = f"{total} = {a} + {b} + {c}"
            return outstr

        expected_result = [
            "@test2_decorator",
            "def test2(a, b, c):",
            '    """A docstring for this test function."""',
            "    # A comment in this test function.",
            "    total = sum(a, b, c)",
            '    outstr = f"{total} = {a} + {b} + {c}"',
            "    return outstr",
        ]

        expected_result = _helper_indent_and_newline(expected_result, 8)

        result = inspect_functions.get_all_source(test2)
        self.assertEqual(result[0], expected_result)

        # Decorators with arguments.
        def test3_decorator(message):
            def test3_decorator(fxn):
                def wrapped_fxn(*args, **kwargs):
                    result = fxn(*args, **kwargs)
                    return f"{message}\n{result}"

                return wrapped_fxn

            return test3_decorator

        @test3_decorator("This was a unit test: ")
        def test3(*vector_values):
            vector = [f"{x:1.3e}" for x in vector_values]
            vectorstr = "".join(vector)

            return vectorstr

        expected_result = [
            '@test3_decorator("This was a unit test: ")',
            "def test3(*vector_values):",
            '    vector = [f"{x:1.3e}" for x in vector_values]',
            '    vectorstr = "".join(vector)',
            "",
            "    return vectorstr",
        ]

        expected_result = _helper_indent_and_newline(expected_result, 8)

        result = inspect_functions.get_all_source(test3)
        self.assertEqual(result[0], expected_result)

        # With multiple decorators.
        def test4_dec1(fxn):
            def wrapper1(*args, **kwargs):
                result = fxn(*args, **kwargs)
                return result

            return wrapper1

        def test4_dec2(fxn):
            def wrapper2(*args, **kwargs):
                result = fxn(*args, **kwargs)
                return result

            return wrapper2

        @test4_dec1
        @test4_dec1
        def test4_1():
            pass

        expected_result = [
            "@test4_dec1",
            "@test4_dec1",
            "def test4_1():",
            "    pass",
        ]

        expected_result = _helper_indent_and_newline(expected_result, 8)
        result = inspect_functions.get_all_source(test4_1)

        self.assertEqual(result[0], expected_result)

        @test4_dec1
        @test4_dec2
        def test4_2():
            pass

        expected_result = [
            "@test4_dec1",
            "@test4_dec2",
            "def test4_2():",
            "    pass",
        ]

        expected_result = _helper_indent_and_newline(expected_result, 8)
        result = inspect_functions.get_all_source(test4_2)

        self.assertEqual(result[0], expected_result)


if __name__ == "__main__":
    unittest.main()



import porchlight
from porchlight import Neighborhood, Door, Param

import unittest
from unittest import TestCase

import logging
import os
import typing

logging.basicConfig(filename=f"{os.getcwd()}/porchlight_unittest.log")


class TestNeighborhood(TestCase):
    def test___init___(self):
        # Initialization should not do anything beyond defining the empty
        # Neighborhood._doors and Neighborhood._params attrs
        neighborhood = Neighborhood()

        self.assertEqual(neighborhood._doors, {})
        self.assertEqual(neighborhood._params, {})

    def test___repr__(self):
        neighborhood = Neighborhood()
        expected = r"Neighborhood(doors={}, params={}, call_order=[])"
        self.assertEqual(repr(neighborhood), expected)

        @Door
        def test1(x: int) -> int:
            y = x + 1
            return y

        neighborhood.add_door(test1)

        expected = (
            "Neighborhood(doors={'test1': Door(name=test1, "
            "base_function=<function test1 at 0x7f247d8c48b0>, "
            "arguments={'x': <class 'int'>}, return_vals=[['y']])}, "
            "params={'x': Param(name=x, value=<porchlight.param."
            "Empty object at 0x7f247e1bd060>, constant=False, type=<class "
            "'porchlight.param.Empty'>), 'y': Param(name="
            "y, value=<porchlight.param.Empty object at 0x7f247d8c9660>, "
            "constant=False, type=<class 'porchlight.param "
            ".Empty'>)}, call_order=['test1'])"
        )

        params = neighborhood.params

        expected = (
            f"Neighborhood(doors={{'test1': {test1}}}, "
            f"params={params}, call_order=['test1'])"
        )

        print(repr(neighborhood))
        self.assertEqual(repr(neighborhood), expected)

    def test_add_function(self):
        # We only need to test if the correct type of door is created---can
        # compare them directly.
        neighborhood = Neighborhood()

        def test1(x, y=1) -> int:
            z = x + y
            return z

        door_equiv = Door(test1)

        neighborhood.add_function(test1)

        self.assertTrue(len(neighborhood._doors) == 1)
        self.assertTrue(len(neighborhood._params) == 3)

        neighborhood_door = list(neighborhood._doors.values())[0]
        self.assertTrue(neighborhood_door == door_equiv)

        # Now try to overwrite the function.
        def test1(x, y=2) -> int:
            z = x + y
            return z

        neighborhood.add_function(test1)

        self.assertEqual(neighborhood._params["y"].value, 1)

        # Now actually overwrite the function.
        neighborhood.add_function(test1, overwrite_defaults=True)

        self.assertEqual(neighborhood._params["y"].value, 2)

    def test_add_door(self):
        # We only need to test if the correct type of door is created---can
        # compare them directly.
        neighborhood = Neighborhood()

        @Door
        def test1(x, y=1) -> int:
            z = x + y
            return z

        neighborhood.add_door(test1)

        self.assertTrue(len(neighborhood._doors) == 1)
        self.assertTrue(len(neighborhood._params) == 3)

        # Now try to overwrite the function.
        @Door
        def test1(x, y=2) -> int:
            z = x + y
            return z

        neighborhood.add_door(test1)

        self.assertEqual(neighborhood._params["y"].value, 1)

        # Now actually overwrite the function.
        neighborhood.add_door(test1, overwrite_defaults=True)

        self.assertEqual(neighborhood._params["y"].value, 2)

        # Add a different Door
        @Door
        def test2(x, z):
            y = x + z
            return y

        neighborhood.add_door(test2)

        self.assertEqual(len(neighborhood._doors), 2)

        # Since Neighborhood.remove_door has only one use case, cover that
        # here.
        neighborhood.remove_door(test2.name)

        self.assertEqual(len(neighborhood._doors), 1)

        neighborhood.remove_door("test1")

        self.assertEqual(len(neighborhood._doors), 0)

        # Testing adding multiple doors as once.
        @Door
        def test1(x, y=1) -> int:
            z = x + y
            return z

        @Door
        def test2(z):
            z = 3 * z
            return z

        @Door
        def test3():
            pass

        @Door
        def test4():
            arbitrary = None
            return arbitrary

        neighborhood = Neighborhood()

        neighborhood.add_door([test1, test2, test3, test4])

        self.assertEqual(len(neighborhood._doors), 4)
        self.assertEqual(len(neighborhood._params), 4)

        # Empty, unused variable test.
        @Door
        def test1(x):
            y = 10 + x
            return y

        @Door
        def test2(x):
            z = 5 + x
            return z

        @Door
        def test3():
            pass

        @Door
        def test4():
            return

        # Mix and match different adding formats.
        neighborhood.add_door(test1)
        neighborhood.add_door(test2)
        neighborhood.add_door([test3, test4])

    @unittest.skip(
        "Already comprehensively tested by other tests for the "
        "time being, but will need to be intdependently tested "
        "once Groups are implemented"
    )
    def test_add_param(self):
        pass

    def test_run_step(self):
        @Door
        def test1(x, y=1) -> int:
            z = x + y
            return z

        @Door
        def test2(z):
            z = 3 * z
            return z

        @Door
        def test3():
            pass

        @Door
        def test4():
            arbitrary = None
            something_else = 5
            return arbitrary, something_else

        neighborhood = Neighborhood()

        neighborhood.add_door([test1, test2, test3, test4])

        neighborhood.add_param("x", 1)
        neighborhood.add_param("z", 1)

        neighborhood.run_step()
        neighborhood.run_step()
        neighborhood.run_step()

        # Check that the state of the parameters is the expected values
        x = neighborhood._params["x"]
        y = neighborhood._params["y"]
        z = neighborhood._params["z"]
        arb = neighborhood._params["arbitrary"]
        smthn = neighborhood._params["something_else"]

        self.assertEqual(x._value, 1)
        self.assertEqual(y._value, 1)
        self.assertEqual(z._value, 6)
        self.assertEqual(smthn._value, 5)
        self.assertTrue(arb._value is None)

        # Test updating a new, empty parameter that will be updated by a return
        # value.
        def retval_empty_test(x):
            rettest = x + 1
            return rettest

        neighborhood.add_function(retval_empty_test)

        self.assertEqual(
            neighborhood._params["rettest"].value, porchlight.param.Empty()
        )

        neighborhood.run_step()

        expected_param = porchlight.param.Param(
            "rettest", neighborhood.params["x"].value + 1
        )

        self.assertEqual(
            neighborhood._params["rettest"].value, expected_param.value
        )

        # Make one of the parameters constant.
        neighborhood.set_param("rettest", 1, constant=True)

        with self.assertRaises(porchlight.param.ParameterError):
            neighborhood.run_step()

    def test_empty_variable_check(self):
        # Empty, unused variable test.
        @Door
        def test1(x):
            y = 10 + x
            return y

        @Door
        def test2(x):
            z = 5 + x
            return z

        @Door
        def test3():
            pass

        @Door
        def test4():
            return

        neighborhood = Neighborhood()
        neighborhood.add_door(test1)
        neighborhood.add_door(test2)
        neighborhood.add_door([test3, test4])

        # Don't pass the required parameters
        with self.assertRaises(porchlight.param.ParameterError):
            neighborhood.run_step()

    def test_order_doors(self):
        neighborhood = Neighborhood()

        @Door
        def test1(x, y, z=1):
            x += y + z
            return x

        @Door
        def test2(z):
            z *= 5
            return z

        @Door
        def test3():
            pass

        @Door
        def test4(x):
            outstr = str(f"My x is: {x}")
            return outstr

        neighborhood.add_door([test1, test2, test3, test4])

        self.assertEqual(
            neighborhood._call_order, ["test1", "test2", "test3", "test4"]
        )

        neighborhood.order_doors(["test2", "test1", "test4", "test3"])

        self.assertEqual(
            neighborhood._call_order, ["test2", "test1", "test4", "test3"]
        )

        # Should raise an error if a door is missing/extra door provided.
        with self.assertRaises(ValueError):
            neighborhood.order_doors(["t", "t", "t", "t", "test4"])

        with self.assertRaises(ValueError):
            neighborhood.order_doors([])

        with self.assertRaises(KeyError):
            neighborhood.order_doors([""])

        with self.assertRaises(KeyError):
            # test2 -> tset2
            neighborhood.order_doors(["test1", "tset2", "test3", "test4"])

    def test_remove_door(self):
        @Door
        def test1(x, y, z=1):
            x += y + z
            return x

        @Door
        def test2(z):
            z *= 5
            return z

        @Door
        def test3():
            pass

        @Door
        def test4(x):
            outstr = str(f"My x is: {x}")
            return outstr

        neighborhood = Neighborhood()
        neighborhood.add_door([test1, test2, test3, test4])

        neighborhood.remove_door("test1")

        self.assertEqual(len(neighborhood.doors), 3)
        self.assertEqual(
            neighborhood._doors,
            {"test2": test2, "test3": test3, "test4": test4},
        )

        with self.assertRaises(KeyError):
            neighborhood.remove_door("bad door")

        # Make sure that the remove door has stayed removed, and that it's no
        # longer referenced in the Neighborhood._doors attr
        with self.assertRaises(KeyError):
            neighborhood.remove_door("test1")

    def test_required_args_present(self):
        @Door
        def test1(x, y=2, *, z):
            return x

        neighborhood = Neighborhood()

        neighborhood.add_door(test1)

        self.assertFalse(neighborhood.required_args_present())

        neighborhood.set_param("x", 7)
        neighborhood.set_param("z", "fourty-two")

        def test2(x, z):
            y = x + z
            return y

        neighborhood.add_function(test2)

        self.assertTrue(neighborhood.required_args_present())

    def test_set_param(self):
        @Door
        def test1(x):
            return

        neighborhood = Neighborhood()

        neighborhood.add_door(test1)

        self.assertEqual(
            neighborhood._params["x"].value, porchlight.param.Empty()
        )

        neighborhood.set_param("x", 1)

        self.assertEqual(neighborhood._params["x"].value, 1)

        # Test constants
        @Door
        def test2(y):
            out = str(y)
            return out

        neighborhood.add_door(test2)

        neighborhood.add_param("y", 1, constant=True)

        with self.assertRaises(porchlight.param.ParameterError):
            neighborhood.set_param("y", 5)

        neighborhood.set_param("y", 5, ignore_constant=True)

        expected_param = porchlight.param.Param("y", 5)
        self.assertEqual(neighborhood._params["y"], expected_param)

    def test_anticipated_parameters(self):
        def fxn_one(x):
            y = x + 1
            return y

        def fxn_two(y):
            return "Hello!"

        neighborhood = Neighborhood()

        neighborhood.add_function(fxn_one)
        neighborhood.add_function(fxn_two)

        # Provide the required first arg, x
        neighborhood.set_param("x", 0)

        neighborhood.run_step()

        # Check that y is now sucessfully initialized.
        expected_param = Param("y", 1)
        self.assertEqual(neighborhood._params["y"], expected_param)

        # Make sure the value doesn't change after another run.
        neighborhood.run_step()
        self.assertEqual(neighborhood._params["y"], expected_param)

        # If another uninitialized value is added but will not be created, the
        # neighborhood should raise an error.
        @Door
        def fxn_three(barf):
            return

        neighborhood.add_door(fxn_three)

        with self.assertRaises(porchlight.param.ParameterError):
            neighborhood.run_step()

    def test_uninitialized_inputs(self):
        @Door
        def test1(x, y=1):
            z = x * y
            return z

        @Door
        def test2(z):
            return

        @Door
        def test3():
            x = 0
            return x

        neighborhood = Neighborhood()

        neighborhood.add_door([test1, test2, test3])

        self.assertEqual(neighborhood.uninitialized_inputs, ["x", "z"])

    def test_dynamic_doors(self):
        # TODO: There is a case for updating multiple doors with one function.
        # It shouldn't be too hard to execute, but need to keep it in mind.
        @Door
        def test1(x: int) -> Door:
            @Door
            def test1_door():
                y = x + 1
                return y

            return test1_door

        neighborhood = Neighborhood()

        # Adding this to the neighborhood should add two doors.
        neighborhood.add_door(test1, dynamic_door=True)

        self.assertEqual(
            list(neighborhood.doors.keys()), ["test1", "test1_door"]
        )

        self.assertEqual(list(neighborhood.params.keys()), ["x", "test1_door"])

        neighborhood.set_param("x", 1)
        neighborhood.run_step()

        self.assertEqual(neighborhood.params["y"].value, 2)

        # Now add a dynamic door with a keyword argument.
        @Door
        def test2() -> Door:
            @Door
            def test2_door(z: int = 3):
                # Note: this will override any previously calculated values of
                # y---on purpose here!
                y = z + 3
                return y

            return test2_door

        neighborhood.add_door(test2, dynamic_door=True)
        neighborhood.run_step()

        self.assertEqual(
            list(neighborhood.doors.keys()),
            ["test1", "test1_door", "test2", "test2_door"],
        )
        self.assertEqual(neighborhood.params["y"].value, 6)
        self.assertEqual(
            list(neighborhood.params.keys()),
            ["x", "test1_door", "y", "test2_door", "z"],
        )

        # This should fail if dynamic doors are not requested.
        neighborhood = Neighborhood()
        neighborhood.add_door(test1)

        self.assertEqual(list(neighborhood.doors.keys()), ["test1"])

        self.assertEqual(list(neighborhood.params.keys()), ["x", "test1_door"])

        # Add two dynamic doors using a single generator function.
        def doublegen_test(
            x: int,
        ) -> typing.Tuple[porchlight.Door, porchlight.Door]:
            @porchlight.Door
            def test1(y: float) -> float:
                z = y ** x + 1
                return z

            @porchlight.Door
            def test2(z: float) -> float:
                x = 2 * z
                return x

            return test1, test2

        neighborhood = Neighborhood()
        neighborhood.add_function(doublegen_test, dynamic_door=True)
        neighborhood.add_param("x", 0)
        neighborhood.add_param("y", 2)

        expected_doors = ["doublegen_test", "test1", "test2"]
        expected_params = ["x", "test1", "test2", "y", "z"]

        self.assertEqual(list(neighborhood.doors.keys()), expected_doors)

        neighborhood.run_step()

        self.assertEqual(list(neighborhood.params.keys()), expected_params)
        self.assertEqual(neighborhood.params["x"].value, 4)
        self.assertEqual(neighborhood.params["y"].value, 2)
        self.assertEqual(neighborhood.params["z"].value, 2)

        neighborhood.run_step()

        self.assertEqual(list(neighborhood.params.keys()), expected_params)
        self.assertEqual(neighborhood.params["x"].value, 34)
        self.assertEqual(neighborhood.params["y"].value, 2)
        self.assertEqual(neighborhood.params["z"].value, 17)

    def test_bad_dynamic_door(self):
        @Door
        def test1(x: int):
            @Door
            def test1_door():
                y = x + 1
                return y

            return test1_door

        neighborhood = Neighborhood()

        # Adding this to the neighborhood should add two doors.
        with self.assertRaises(porchlight.neighborhood.NeighborhoodError):
            neighborhood.add_door(test1, dynamic_door=True)

    def test_properties(self):
        @porchlight.Door
        def test1(x: typing.Union[int, float], y: str = "15.5") -> typing.Any:
            z = x + float(y)
            return z

        neighborhood = Neighborhood()
        neighborhood.add_door(test1)

        # Neighborhood.parameters
        result = neighborhood.parameters
        self.assertEqual(result["x"], Param("x", porchlight.param.Empty()))
        self.assertEqual(result["y"], Param("y", "15.5"))


if __name__ == "__main__":
    import unittest

    unittest.main()



"""Tools for introspection of functions extending what :py:module:`inspect` can
do.
"""
import inspect

from typing import Callable, List, Tuple, Type


def get_all_source(function: Callable) -> Tuple[List[str], int]:
    """Retrieves all source code related to a given function, even if it has
    been otherwise wrapped.

    It returns a tuple containing a list of strings containing the source code
    and an integer (starting line number). This is output by the eventual call
    to `inspect.getsourcelines` on the wrapped function.

    Arguments
    ---------
    function : Callable
        A defined function to get the source code for.
    """
    if not isinstance(function, Callable):
        raise TypeError(
            f"Source lines can only be retrieved for Callable "
            f"objects, not {type(function)}."
        )

    if "__closure__" in dir(function) and function.__closure__:
        # Recursively dive down. the first closure cell value should be the
        # next function down.
        cell = function.__closure__[0].cell_contents

        if isinstance(cell, Callable) and not isinstance(cell, Type):
            return get_all_source(function.__closure__[0].cell_contents)

    sourcelines = inspect.getsourcelines(function)

    return sourcelines


def get_wrapped_function(function: Callable) -> Callable:
    """If the input callable has the `__closure__` attr and its first cell is a
    function, it will descend until it has found a callable object with no
    `__closure__` variable.
    """
    if not isinstance(function, Callable):
        raise TypeError(
            f"Source lines can only be retrieved for Callable "
            f"objects, not {type(function)}."
        )

    if "__closure__" in dir(function) and function.__closure__:
        # Recursively dive down. the first closure cell value should be the
        # next function down.
        cell = function.__closure__[0].cell_contents

        if isinstance(cell, Callable) and not isinstance(cell, Type):
            return get_wrapped_function(function.__closure__[0].cell_contents)

    return function



"""Tests for the porchlight.door.DynamicDoor class."""
import porchlight.door as door

import os
import unittest


class TestDynamicDoor(unittest.TestCase):
    def test___init__(self):
        @door.DynamicDoor
        def doorgen1() -> door.Door:
            @door.Door
            def my_door(x: int) -> int:
                y = x + 1
                return y

            return my_door

        self.assertEqual(doorgen1._door_generator.__name__, "doorgen1")
        self.assertEqual(doorgen1.generator_args, [])
        self.assertEqual(doorgen1.generator_kwargs, {})

        doorgen1.update()

        self.assertEqual(doorgen1.name, "my_door")
        self.assertEqual(doorgen1.variables, ["x", "y"])

        _y = doorgen1(3)

        self.assertEqual(_y, 4)

    def test___repr__(self):
        @door.DynamicDoor
        def doorgen1() -> door.Door:
            @door.Door
            def my_door(x: int) -> int:
                y = x + 1
                return y

            return my_door

        # Initially, should be uninitialized with no meta-info.
        self.assertEqual(repr(doorgen1), "DynamicDoor(uninitialized)")

        # Once the generator function has been run once and the Dynamic door
        # has been updated, expect a door-like string with "Dynamic" prepended.
        doorgen1(1)
        my_door = doorgen1._cur_door
        expected_repr = repr(my_door)
        expected_repr = expected_repr.replace("Door", "DynamicDoor")

        self.assertEqual(repr(doorgen1), expected_repr)

    def test_bad_returns(self):
        # Non-BaseDoor returns should raise a TypeError at update.
        def test1() -> int:
            return 1

        def test2() -> str:
            x = ""
            return x

        def test3(x: int) -> int:
            return x

        def test4() -> door.Door:
            return

        def test5() -> door.BaseDoor:
            x = 1
            return x

        def test6(bing: str) -> door.Door:
            def fxn():
                return "hello"

            return fxn

        all_tests = [test1, test2, test3, test4, test5, test6]

        for test in all_tests:
            with self.assertRaises(TypeError):
                new_ddoor = door.DynamicDoor(test)
                new_ddoor.update()


if __name__ == "__main__":
    unittest.main()



"""Tests to verify the integrity of the porchlight.door.Door class and any
relevant helper functions or objects.
"""
from unittest import TestCase

from porchlight.door import Door
from porchlight.param import Empty, ParameterError

import logging
import os

logging.basicConfig(filename=f"{os.getcwd()}/porchlight_unittest.log")


class TestDoor(TestCase):
    def test___init__(self):
        def test_fxn(x: int) -> int:
            y = 2 * x
            return y

        door = Door(test_fxn)

        # Must contain both input and output parameter.
        arguments = ["x"]
        keyword_args = ["x"]
        return_vals = [["y"]]

        # Not comparing any values during this test.
        for arg in arguments:
            self.assertIn(arg, door.arguments)

        for kwarg in keyword_args:
            self.assertIn(kwarg, door.keyword_args)

        for retval in return_vals:
            self.assertIn(retval, door.return_vals)

        # Call the Door
        result = door(x=5)

        self.assertEqual(result, 10)

        # Decorators (with defaults) should behave as expected
        @Door
        def fxn_use_decorator(x):
            z = 2 * x
            return z

        self.assertEqual(fxn_use_decorator(10), 20)

        self.assertEqual(fxn_use_decorator.name, "fxn_use_decorator")

        self.assertEqual(fxn_use_decorator.arguments, {"x": Empty})

    def test___call__(self):
        def test_fxn(x: int) -> int:
            y = 2 * x
            return y

        door = Door(test_fxn)

        # Call the Door with erroneous types based on annotations.
        with self.assertRaises(ParameterError):
            door(x="6")

    def test_required_arguments(self):
        # This property is critical for the functioning of the Neighborhood
        # object, since it determines what areguments are passed and *how* they
        # are passed.
        @Door
        def test1(x, y=1):
            x += y
            return x

        @Door
        def test2(a, b, *, x):
            b += a * x
            x += a
            return b, a

        @Door
        def test3(key1="beta", key2="beeeeta"):
            return

        @Door
        def test4():
            pass

        @Door
        def test5():
            x = 5
            return x

        test_cases = (
            (test1, ["x"]),
            (test2, ["a", "b", "x"]),
            (test3, []),
            (test4, []),
            (test5, []),
        )

        for fxn, exp in test_cases:
            self.assertEqual(fxn.required_arguments, exp)

    def test_variables(self):
        @Door
        def test1(x, y=1):
            x += y
            return x

        @Door
        def test2(a, b, *, x):
            b += a * x
            x += a
            return b, a

        @Door
        def test3(key1="beta", key2="beeeeta"):
            return

        @Door
        def test4():
            pass

        @Door
        def test5():
            x = 5
            return x

        test_cases = (
            (test1, ["x", "y"]),
            (test2, ["a", "b", "x"]),
            (test3, ["key1", "key2"]),
            (test4, []),
            (test5, ["x"]),
        )

        for fxn, exp in test_cases:
            self.assertEqual(fxn.variables, exp)

    def test___repr___(self):
        def test(x: int) -> int:
            y = x ** 2
            return y

        expected_repr = (
            f"Door(name=test, base_function={str(test)}, "
            f"arguments={{'x': <class 'int'>}}, "
            f"return_vals=[['y']])"
        )

        test_door = Door(test)
        self.assertEqual(repr(test_door), expected_repr)


if __name__ == "__main__":
    import unittest

    unittest.main()



"""Tests for functions contained in ../utils/typing_functions."""
from porchlight.utils import typing_functions
from typing import Callable, Dict, Tuple

import unittest


class TestTypingFunctions(unittest.TestCase):
    def test_decompose_type(self):
        # TODO: The below should be reorganized to have tests go from defaults
        # -> messing with parameters. Right now, confusing that
        # include_base_types == True isn't covered until after other test
        # cases.
        # Define test cases as input/expected output pairs.
        test_cases = (
            ([str], [str]),
            ([Tuple], [Tuple]),
            ([Tuple[str]], [str]),
            ([Dict[str, list]], [str, list]),
            ([Tuple[Callable[str, float], str]], [Callable[str, float], str]),
        )

        for test in test_cases:
            decomposition = typing_functions.decompose_type(
                *test[0], [], False
            )

            self.assertEqual(decomposition, test[1])

        # Examples with break_types != []
        test_cases = (
            ([str, [float]], [str]),
            ([Tuple, [Tuple]], [Tuple]),
            ([Tuple[str], [Tuple]], [Tuple[str]]),
            ([Dict[str, list[str]], [list]], [str, list[str]]),
            ([Callable[str, float], [Callable]], [Callable[str, float]]),
        )

        for test in test_cases:
            decomposition = typing_functions.decompose_type(*test[0], False)

            self.assertEqual(decomposition, test[1])

        # Examples with include_base_types = True
        test_cases = (
            ([str, []], [str]),
            ([Tuple, []], [Tuple]),
            ([Tuple[str], []], [Tuple[str], str]),
            (
                [Dict[str, list[str]], []],
                [Dict[str, list[str]], str, list[str]],
            ),
            ([Callable[str, float], []], [Callable[str, float]]),
        )

        for test in test_cases:
            decomposition = typing_functions.decompose_type(*test[0])

            self.assertEqual(decomposition, test[1])


if __name__ == "__main__":
    # Run test
    unittest.main()
